; _______________________________________________________________________________________
;|                                                                                       |
;| ..::FreshLib::..  Free, open source. Licensed under "BSD 2-clause" license."          |
;|_______________________________________________________________________________________|
;
;  Description: Contains the base procedures for the FreshLib OOP framework
;
;  Target OS: Any
;
;  Dependencies: memory.asm
;
;  Notes:
;
;  TODO: Thread safety! The whole FreshLib should be thread safe. GUI part,as long as
;        uses memory objects should be thread safe. This can be achieved by locking
;        objects, during processing. objects.asm is the proper place for such lock mechanism.
;        The lock should be provided by some kind of mutex objects, associated with the objects.
;
;_________________________________________________________________________________________
module "Objects library"

struct TObjectClass
  .ptrParent        dd 0   ; pointer to the parent TObjectClass
  .dataSize         dd 0   ; size of TObject structure.
  .procCreate       dd 0
  .procDestroy      dd 0
  .procGetParam     dd 0
  .procSetParam     dd 0
  .procExecCmd      dd 0   ; executes method in chain.
  .procSysEventHandler dd 0
ends



macro ObjectClass name*, parent*, procCreate*, procDestroy*, procGetParam*, procSetParam*,  procExecCmd*, procSysEvents* {
  if used C#name
    C#name TObjectClass
    store dword C#parent at C#name#.ptrParent
    store dword sizeof.T#name at C#name#.dataSize
    store dword procCreate at C#name#.procCreate
    store dword procDestroy at C#name#.procDestroy
    store dword procGetParam at C#name#.procGetParam
    store dword procSetParam at C#name#.procSetParam
    store dword procExecCmd at C#name#.procExecCmd
    store dword procSysEvents at C#name#.procSysEventHandler
  end if
}


CRoot = 0


maskParameter = $80000000
Sequence param, maskParameter


macro method name*, [arg] {
common
 local ..argcount
 ..argcount = 0
   param name
 forward
   if ~arg eq
   ..argcount = ..argcount + 1
   end if
 common
   name#.argcount = ..argcount
}


macro execute obj*, meth*, [arg] {
reverse
  if ~arg eq
    pushx arg
  end if
common
  pushd meth#.argcount
  pushd meth
  pushd obj
  call  __Exec
}


macro object name*, parent {
  macro name@object \{ name name \}
  macro size@object \{ sizeof.#name = $ \}
  struc name \{
    param.current = maskParameter
    if ~parent eq
      . parent
    else
      .:
    end if
}

endobj fix } obj_helper

macro obj_helper {
  virtual at 0
    name@object
    size@object
  end virtual
  purge name@object
  purge size@object
}


mxTimeout = 1000        ; 1s timeout for the access mutex objects.

; common object procedures
; These procedures work with every TObject descendent object type.


;_________________________________________________________________________________________
;
; proc Create, .class - creates an instance from the given class.
;
; Arguments:
;   .class - pointer to TObjectClass structure for needed object.
;
; Returns:
;   ebx = pointer to the instance of the object.
;
; This procedure calls procCreate for the given object class and all
; parent classes in reverce order.
;_________________________________________________________________________________________

proc Create, .class
begin
        push    eax ecx

        mov     ecx, [.class]
        stdcall GetMem, [ecx+TObjectClass.dataSize]
        mov     ebx, eax
        mov     [eax+TObject.ptrClass], ecx

  if options.Threads
        lea     eax, [eax+TObject.mxAccess]
        stdcall MutexCreate, 0, eax
  end if

        push    0

.loop:
        cmp     [ecx+TObjectClass.procCreate], 0
        je      .parent

        push    [ecx+TObjectClass.procCreate]

.parent:
        mov     ecx, [ecx+TObjectClass.ptrParent]
        test    ecx, ecx
        jnz     .loop

.loop2:
        pop     ecx
        jecxz   .endcreate

        push    ebx
        stdcall ecx, ebx
        pop     ebx
        jmp     .loop2

.endcreate:

  if options.Threads
        lea     eax, [ebx+TObject.mxAccess]
        stdcall MutexRelease, eax
  end if

        pop     ecx eax
        return
endp



;_________________________________________________________________________________________
;
; proc Destroy, .obj - destroys the object instance.
;
; Arguments:
;   .obj - pointer to TObject descendent object.
;
; Returns:
;   nothing.
;
; This procedure calls procDestroy for the given object class and all
; parent classes. Note that the memory is freed only by the last
; destroy handled from CObject class.
;_________________________________________________________________________________________

proc Destroy, .obj
begin
        push    eax ebx

        mov     ebx, [.obj]
        test    ebx, ebx
        jz      .error

  if options.Threads
        stdcall _LockObject, [.obj]
  end if

        mov     ebx, [ebx+TObject.ptrClass]

.loop:
        cmp     [ebx+TObjectClass.procDestroy], 0
        je      .parent

        mov      eax, [ebx+TObjectClass.procDestroy]

        push    ebx
        stdcall [ebx+TObjectClass.procDestroy], [.obj]
        pop     ebx

.parent:
        mov     ebx, [ebx+TObjectClass.ptrParent]
        test    ebx, ebx
        jnz     .loop

        mov     ebx, [.obj]

  if options.Threads
        lea     eax, [ebx+TObject.mxAccess]
  end if

        mov     ebx, [ebx+TObject.ptrVar]
        test    ebx, ebx
        jz      @f
        mov     dword [ebx], 0
@@:
  if options.Threads
        stdcall MutexDestroy, eax
  end if

        stdcall FreeMem, [.obj]
        clc
        pop     ebx eax
        return

.error:
  if options.Threads
        stdcall _UnlockObject, [.obj]
  end if
        stc
        pop     ebx eax
        return
endp




;_________________________________________________________________________________________
;
; proc Get, .obj, .paramID - returns the value of given parameter
;
; Arguments:
;   .obj - pointer to TObject descendent object.
;   .paramID - ID number of the parameter being read.
; Returns:
;   CF=0; eax = parameter value.
;   CF=1; the parameter is unknown and can not be retrived.
;
; This procedure calls procGet for the given object class and
; if the procedure returns CF=1; tries to call the parents procedure
; procGet, until the value is retrived, or the root level of inheritance
; is reached.
;_________________________________________________________________________________________

proc Get, .obj, .paramID
begin
        push    ecx

  if options.Threads
        stdcall _LockObject, [.obj]
  end if

        mov     ecx, [.obj]
.loop0:
        stc
        jecxz   .finish

.loop:
        mov     ecx, [ecx]      ; TObject.ptrClass = TObjectClass.ptrParent = 0
        jecxz   .finish

        cmp     [ecx+TObjectClass.procGetParam], 0
        je      .loop0

        push    ecx
        stdcall [ecx+TObjectClass.procGetParam], [.obj], [.paramID]
        pop     ecx
        jc      .loop

.finish:
  if options.Threads
        stdcall _UnlockObject, [.obj]
  end if
        pop     ecx
        return
endp



;_________________________________________________________________________________________
;
; proc Set, .obj, .paramID - sets the value of given parameter
;
; Arguments:
;   .obj - pointer to TObject descendent object.
;   .paramID - ID number of the parameter being read.
;   .value   - value of the parameter that should be set.
; Returns:
;   CF=0 - the value is successfully set.
;   CF=1; the value is not set, because the parameter is unknown.
;
; This procedure calls procSet for the given object class and
; if the procedure returns CF=1; tries to call the parents procedures
; procSet, until the value is set, or the root level of inheritance
; is reached.
;_________________________________________________________________________________________

proc Set, .obj, .paramID, .value
begin
        push    ecx

  if options.Threads
        stdcall _LockObject, [.obj]
  end if
        mov     ecx, [.obj]
.loop0:
        stc
        jecxz   .finish

.loop:
        mov     ecx, [ecx]      ; TObject.ptrClass = TObjectClass.ptrParent = 0
        jecxz   .finish

        cmp     [ecx+TObjectClass.procSetParam], 0
        je      .loop0

        push    ecx
        stdcall [ecx+TObjectClass.procSetParam], [.obj], [.paramID], [.value]
        pop     ecx
        jc      .loop

.finish:
  if options.Threads
        stdcall _UnlockObject, [.obj]
  end if

        pop     ecx
        return
endp





;_________________________________________________________________________________________
;
; proc ExecEvent, .obj, .event
; Arguments:
;   .obj - pointer to the TObject descendent object.
;   .event - pointer to TSysEvent descendent structure.
;_________________________________________________________________________________________

proc ExecEvent, .obj, .event
begin
        push    ecx

  if options.Threads
        stdcall _LockObject, [.obj]
  end if
        mov     ecx, [.obj]
.loop0:
        stc
        jecxz   .finish

.loop:
        mov     ecx, [ecx]      ; TObject.ptrClass = TObjectClass.ptrParent = 0
        jecxz   .finish

        cmp     [ecx+TObjectClass.procSysEventHandler], 0
        je      .loop0

        push    ecx
        stdcall [ecx+TObjectClass.procSysEventHandler], [.obj], [.event]
        pop     ecx
        jc      .loop

.finish:
  if options.Threads
        stdcall _UnlockObject, [.obj]
  end if
        pop     ecx
        return
endp




;_________________________________________________________________________________________
;
; proc IsObject, .class - returns CF=0 if the given object belongs to
;                         the given class. Else returns CF=1
;_________________________________________________________________________________________
proc IsObject   ; , .obj, .class
virtual at esp+8
  .obj dd ?
  .class dd ?
end virtual
begin
        push    ecx
        lea     ecx, [.obj]

.loop:
        stc
        mov     ecx, [ecx]      ; TObject.ptrClass = TObjectClass.ptrParent = 0
        jecxz   .finish

        cmp     ecx, [.class]   ; the first cmp will always fail because ecx contains pointer to obj that can't be equal to a class.
        jne     .loop

.finish:
        pop     ecx
        retn 8
endp



;_________________________________________________________________________________________
;
; Executes a method of given class.
; This procedure have arbitrary count of arguments
; it calls in chain procExecCmd the classes and passes
; a pointer to the arguments in esi.
; It first calls the procExecCmd of the childs and then of the parents.
; If some class have 0 in [procExecCmd] this class is skiped and
; the execution continues on its parent.
; If there is no proper handler back to the root, then the procedure
; returns CF=1
;_________________________________________________________________________________________

proc __Exec ;, .obj, .method, .argcount, ..
begin
virtual at ebp+8
  .obj dd ?
  .method dd ?
  .argcount dd ?
  .arguments:
end virtual
        push    ebp
        mov     ebp, esp        ; now ebp+8 points to the arguments.

        push    ecx ebx

        xor     ebx, ebx
        cmp     [.argcount], 0
        je      @f
        lea     ebx, [.arguments]
@@:
        mov     ecx, [.obj]

.loop0:
        stc
        jecxz   .exit

.loop:
        mov     ecx, [ecx]      ; TObject.ptrClass = TObjectClass.ptrParent = 0
        jecxz   .exit

        cmp     [ecx+TObjectClass.procExecCmd], 0
        je      .loop0

        push    ecx ebx
        stdcall [ecx+TObjectClass.procExecCmd], [.obj], [.method]
        pop     ebx ecx
        jc      .loop

.exit:
        pop     ebx ecx

        ; experimental return from procedure with arbitrary count of arguments.
        ; with keeping all registers unchanged.
        mov     esp, ebp
        push    ecx
        mov     ebp, [.argcount]
        lea     ebp, [esp+8+4*ebp+12]      ; the place where the return address should be placed. (Exec have 3 more arguments, including .argcount)
        mov     ecx, [esp+8]               ; return address
        mov     [ebp], ecx                 ; move the return address to the proper place.
        pop     ecx                        ; restore ecx
        pop     esp                        ; old value of ebp
        xchg    esp, ebp                   ; restore ebp and set esp to the proper value
        retn
endp



if options.Threads

  proc _LockObject, .obj
  begin
          push    eax

          mov     eax, [.obj]

          lea     eax, [eax+TObject.mxAccess]
          stdcall WaitForMutex, eax, mxTimeout
          jc      .stop

          pop     eax
          return

  .stop:
          int3
          pop     eax
          return

  endp


proc _UnlockObject, .obj
begin
        pushf
        push    eax

        mov     eax, [.obj]
        lea     eax, [eax+TObject.mxAccess]
        stdcall MutexRelease, eax

        pop     eax
        popf
        return
endp

end if




endmodule
